// Copyright (c) 2023 VMware Carbon Black, all rights reserved

// Package app -- Methods and classes central to the app itself
package app

import (
	"fmt"
	"log"
	"os"
	"os/signal"
	"runtime"
	"runtime/debug"
	"strings"
	"sync/atomic"
	"syscall"

	"gitlab.bit9.local/psc/ems/hinting/internal/config"
	"gitlab.bit9.local/psc/ems/hinting/internal/dbwrapper"
	"gitlab.bit9.local/psc/ems/hinting/internal/logger"
	"gitlab.bit9.local/psc/ems/hinting/internal/metrics"
	"gitlab.bit9.local/psc/ems/hinting/internal/redisclient"
	"gitlab.bit9.local/psc/ems/hinting/internal/utils"

	"go.uber.org/zap"
)

// The HintingService app, where all the magic happens

// HintingService -- Contains all the components required to execute the service
type HintingService struct {
	cfg           *config.Config
	HttpServer    *HttpServer
	dbClient      *dbwrapper.DbClient
	redisClient   redisclient.RedisClient
	metricsClient metrics.Service
	alive         *int32
	Ready         *int32
	Shutdown      *utils.SyncedCondBool
	signals       <-chan os.Signal
	Logger        logger.Logger
}

// ----- Begin Service methods --------------------

// NewHintingService -- Create a new hinting service from a default configuration.
func NewHintingService() *HintingService {
	return NewHintingServiceFromCfg(config.GetInstance(), hookSignals())
}

// NewHintingServiceFromCfg -- Create a new hinting service from a supplied configuration.
func NewHintingServiceFromCfg(cfg *config.Config, signals <-chan os.Signal) *HintingService {
	return &HintingService{Logger: logger.BaseLogger(), cfg: cfg, Shutdown: utils.NewSyncBool(false), Ready: new(int32), alive: new(int32), signals: signals}
}

func (service *HintingService) setupAndRun() error {
	metricsClient, wfErr := metrics.NewClient(service.cfg)
	if wfErr != nil {
		logger.BaseLogger().Warn("Unable to configure wavefront metrics client, metris support will be disabled", zap.Error(wfErr))
	}

	logLevel := config.GetLoggerConfigInstance().LogLevel
	if logLevel == "debug" && metricsClient != nil {
		metrics.RegisterGoStatMetrics(metricsClient)
	}
	service.metricsClient = metricsClient

	dbClient, err := dbwrapper.NewDbClientWithMetrics(service.cfg, service.alive, service.Ready, service.metricsClient)
	if err != nil {
		return err
	}
	service.dbClient = dbClient

	redisClient, err := redisclient.NewRedisClientWithTimeoutFromConfig(service.cfg)
	if err != nil {
		return err
	}
	service.redisClient = redisClient

	service.HttpServer, err = NewHttpServer(service)
	if err != nil {
		return err
	}
	logger.BaseLogger().Info(fmt.Sprintf("Golang runtime version: %v", runtime.Version()))
	logger.BaseLogger().Info(fmt.Sprintf("Hinting-api for %s listening on %s\n", service.cfg.CbdEnvName, service.cfg.ServerAddr))
	logger.BaseLogger().Info(fmt.Sprintf("Log level: %v", logLevel))
	atomic.StoreInt32(service.alive, 1)
	atomic.StoreInt32(service.Ready, 1)
	return service.runUntilToldToShutdown()
}

func hookSignals() <-chan os.Signal {
	quit := make(chan os.Signal, 100)
	signal.Notify(quit, syscall.SIGINT, syscall.SIGTERM)
	return quit
}

func (service *HintingService) runUntilToldToShutdown() error {
	<-service.signals
	return service.shutdown()
}

func panicLogger(desc string) {
	if r := recover(); r != nil {
		logger.BaseLogger().Error(desc, zap.Error(fmt.Errorf("%v", r)))
	}
}

func (service *HintingService) closeDb() error {
	defer panicLogger("Panic closing db")
	if service.dbClient != nil {
		return service.dbClient.Close()
	}
	return nil
}

func (service *HintingService) closeRedis() {
	defer panicLogger("Panic closing redis")
	if service.redisClient != nil {
		service.redisClient.Close()
	}
}

func (service *HintingService) closeHttpServer() error {
	defer panicLogger("Panic closing httpserver")
	if service.HttpServer != nil {
		return service.HttpServer.Shutdown()
	}
	return nil
}

func (service *HintingService) shutdown() error {
	logger.BaseLogger().Info("Shutting down server...")

	httpCloseErr := service.closeHttpServer()

	service.closeRedis()

	dbCloseErr := service.closeDb()

	if service.metricsClient != nil {
		service.metricsClient.CloseConnection()
	}

	errors := make([]error, 3)
	errCount := 0
	if httpCloseErr != nil {
		errCount++
		errors[0] = fmt.Errorf("error closing http server %w: ", httpCloseErr)
	}
	if dbCloseErr != nil {
		errCount++
		errors[2] = fmt.Errorf("error closing db connection %w: ", dbCloseErr)
	}
	if errCount > 0 {
		errStrings := make([]string, errCount)
		errStringsIndex := 0
		for _, err := range errors {
			if err != nil {
				errStrings[errStringsIndex] = err.Error()
				errStringsIndex++
			}
		}
		return fmt.Errorf("shutdown encountered error(s): \n%v", strings.Join(errStrings, "\n"))
	}

	return nil
}

// RunForever -- Main loop of the service
func (service *HintingService) RunForever() {
	defer func() {
		err := recover()
		if err != nil {
			logger.BaseLogger().Error("HintingService exiting from error", zap.Any("error", err), zap.String("stacktrace", string(debug.Stack())))
		}
	}()

	if err := service.setupAndRun(); err != nil {
		log.Printf("Service excited with error:\n %v \n", err)
	}

	service.Shutdown.Set(true)
	log.Println("Service exited gracefully")
}
