// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package controllers

import (
	"errors"
	"fmt"
	"net/http"

	"gitlab.bit9.local/psc/ems/hinting/internal/dbwrapper"
	"gitlab.bit9.local/psc/ems/hinting/internal/redisclient"

	"github.com/gin-gonic/gin"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/middleware"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/models"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/repos"
	"go.uber.org/zap"
)

// Controller interface
type Controller interface {
	Update(*gin.Context)
}

// GenericController -- generic repository structure used to handle device/group/org/env models
type GenericController[DBModel models.Model, APIModel any, PointerAPIModel models.APIModelPointer[APIModel, DBModel]] struct {
	repo repos.Repository[DBModel]
}

// NewController -- function used to allocate a controller for each configuration model
func NewController[DBModel models.Model, APIModel any, PointerAPIModel models.APIModelPointer[APIModel, DBModel]](db *dbwrapper.DbClient, redisClient redisclient.RedisClient) Controller {
	return GenericController[DBModel, APIModel, PointerAPIModel]{repo: repos.NewCachedRepoFromRepo[DBModel](redisClient, repos.NewRepository[DBModel](db))}
}

// //////////////////////////////////////////////////////////////////////////////
// Controller function implementations
// //////////////////////////////////////////////////////////////////////////////

// Update -- Update a database object with supplied data object
func (controller GenericController[DBModel, APIModel, PointerAPIModel]) Update(c *gin.Context) {
	var model APIModel
	var apiModelPointer PointerAPIModel = &model

	if err := c.ShouldBindUri(apiModelPointer.GetURIPart()); err != nil {
		_ = c.AbortWithError(http.StatusBadRequest, fmt.Errorf("failed to parse uri parameters: %w", err))
		return
	}

	if err := c.ShouldBindJSON(apiModelPointer.GetBodyPart()); err != nil {
		_ = c.AbortWithError(http.StatusBadRequest, fmt.Errorf("failed to parse body: %w", err))
		return
	}

	dbModel, err := apiModelPointer.ConvertToDBObject()
	if err != nil {
		if errors.Is(err, models.ErrModelConversionWarning) {
			_ = c.Error(err)
		} else {
			_ = c.AbortWithError(http.StatusBadRequest, fmt.Errorf("failed to convert to database model: %w", err))
			return
		}
	}

	c.Set(string(middleware.ExtraLogFields), []zap.Field{zap.Inline(dbModel)})

	// NOTE: Currently unable to generate coverage for this
	if err := controller.repo.Update(c.Request.Context(), dbModel); err != nil {
		_ = c.AbortWithError(http.StatusInternalServerError, fmt.Errorf("failed to update database entry: %w", err))
		return
	}
}
