// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package controllers

import (
	"bytes"
	"encoding/json"
	"fmt"
	"net/http"
	"net/http/httptest"
	"testing"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/devicechangetype"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/devicegroupchangetype"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/orgchangetype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/deploymenttype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/devicetype"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/repos"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

////////////////////////////////////////////////////////////////////////////////
// Test helper definitions
////////////////////////////////////////////////////////////////////////////////

// sampleDeviceBody -- a sample APIDeviceBody with one or more Changes, predefined as WINDOWS type and ENDPOINT deployment
func sampleDeviceBody(changes ...string) *models.APIDeviceBody {
	return &models.APIDeviceBody{
		Changes: changes,
	}
}

// sampleDeviceGroupBody -- a sample APIDeviceGroupBody with one or more Changes
func sampleDeviceGroupBody(changes ...string) *models.APIDeviceGroupBody {
	return &models.APIDeviceGroupBody{
		Changes: changes,
	}
}

// sampleOrgBody -- a sample APIOrgBody with one or more Changes, predefined with WIDOWS device type
func sampleOrgBody(isKillSwitchOn bool) *models.APIOrgBody {
	deviceTypes := []string{string(devicetype.WINDOWS)}
	changes := []string{orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()}
	deploymentChanges := map[string][]string{string(deploymenttype.AWS): {orgchangetype.ORG_DISABLE_DEFENSE_RULES.String()}}

	return &models.APIOrgBody{
		CanaryPayloadDeviceTypes: deviceTypes,
		IsKillSwitchOn:           &isKillSwitchOn,
		Changes:                  changes,
		DeploymentTypeChanges:    deploymentChanges,
	}
}

// sampleEnvBody -- a sample APIEnvBody with one or more CanaryPayloadDeviceTypes
func sampleEnvBody(certServiceTimeChanged, policyServiceUpdateTimeChanged bool, deviceTypes ...string) *models.APIEnvBody {
	return &models.APIEnvBody{
		CertServiceTimeChanged:   certServiceTimeChanged,
		PolicyServiceTimeChanged: policyServiceUpdateTimeChanged,
		CanaryPayloadDeviceTypes: deviceTypes,
	}
}

////////////////////////////////////////////////////////////////////////////////
// Begin Tests
////////////////////////////////////////////////////////////////////////////////

var _ = Describe("Controller", func() {
	controllerDetails := []struct {
		name       string
		uri        string
		controller Controller
		model      any
		params     []gin.Param
		badParams  []gin.Param
	}{
		{
			"Device",
			"orgs/N86QADZN/devices/1337",
			GenericController[*models.Device, models.APIDevice, *models.APIDevice]{repo: repos.NewMockRepository[models.Device]()},
			sampleDeviceBody(devicechangetype.DEV_GROUP_CHANGE.String()),
			[]gin.Param{
				{Key: "deviceId", Value: "1337"},
				{Key: "orgKey", Value: orgKey},
			},
			[]gin.Param{
				{Key: "deviceId", Value: "1337"},
				{Key: "orgKey", Value: "NOTANORGKEY"},
			},
		},
		{
			"Device Group",
			"orgs/N86QADZN/devicegroups/1337",
			GenericController[*models.DeviceGroup, models.APIDeviceGroup, *models.APIDeviceGroup]{repo: repos.NewMockRepository[models.DeviceGroup]()},
			sampleDeviceGroupBody(devicegroupchangetype.DG_LIVE_QUERY_CHANGE.String(), devicegroupchangetype.DG_POLICY_CHANGE.String()),
			[]gin.Param{
				{Key: "deviceGroupId", Value: "1337"},
				{Key: "orgKey", Value: orgKey},
			},
			[]gin.Param{
				{Key: "deviceGroupId", Value: "1337"},
				{Key: "orgKey", Value: "NOTANORGKEY"},
			},
		},

		{
			"Organization",
			"orgs/N86QADZN",
			GenericController[*models.Organization, models.APIOrg, *models.APIOrg]{repo: repos.NewMockRepository[models.Organization]()},
			sampleOrgBody(true),
			[]gin.Param{
				{Key: "orgKey", Value: orgKey},
			},
			[]gin.Param{
				{Key: "orgKey", Value: "NOTANORGKEY"},
			},
		},

		{
			"Environment",
			"name/QA01",
			GenericController[*models.Environment, models.APIEnv, *models.APIEnv]{repo: repos.NewMockRepository[models.Environment]()},
			sampleEnvBody(true, true,
				string(devicetype.IOS), string(devicetype.WINDOWS)),
			[]gin.Param{
				{Key: "envName", Value: "QA01"},
			},
			[]gin.Param{},
		},
	}

	for _, details := range controllerDetails {
		Context(fmt.Sprintf("Controller %s tests", details.name), func() {
			uri := details.uri
			controller := details.controller
			model := details.model
			data, _ := json.Marshal(model)
			params := details.params
			badParams := details.badParams

			testConfigs := []struct {
				HTTPMethod       string
				HTTPSuccessCode  int
				ControllerMethod func(*gin.Context)
				requiresBody     bool
			}{
				{HTTPMethod: http.MethodPut, HTTPSuccessCode: 200, ControllerMethod: controller.Update, requiresBody: true},
			}
			for _, testConfig := range testConfigs {
				Context(fmt.Sprintf("HTTP %s tests", testConfig.HTTPMethod), func() {
					fmt.Printf("Method: %s %s\n", testConfig.HTTPMethod, uri)
					verb := testConfig.HTTPMethod
					controllerMethod := testConfig.ControllerMethod
					httpSuccessCode := testConfig.HTTPSuccessCode
					requiresBody := testConfig.requiresBody
					var body *bytes.Buffer
					if requiresBody {
						body = bytes.NewBuffer(data)
					} else {
						body = bytes.NewBufferString("")
					}

					It("Accepts valid model input", func() {
						context, _ := gin.CreateTestContext(httptest.NewRecorder())
						context.Request, _ = http.NewRequest(verb, uri, body)
						context.Params = params
						controllerMethod(context)
						Expect(context.Writer.Status()).To(BeEquivalentTo(httpSuccessCode))
					})

					if requiresBody {
						It("Does not accept invalid model input", func() {
							context, _ := gin.CreateTestContext(httptest.NewRecorder())
							context.Request, _ = http.NewRequest(verb, uri, http.NoBody)
							context.Params = params
							controllerMethod(context)
							Expect(context.Writer.Status()).To(BeEquivalentTo(http.StatusBadRequest))
						})
						if len(details.badParams) > 0 {
							It("Does not accept invalid params (eg, orgkey being invalid) ", func() {
								context, _ := gin.CreateTestContext(httptest.NewRecorder())
								context.Request, _ = http.NewRequest(verb, uri, http.NoBody)
								context.Params = badParams
								controllerMethod(context)
								Expect(context.Writer.Status()).To(BeEquivalentTo(http.StatusBadRequest))
							})
						}
					}
				})
			}
		})
	}

	// ----- Additional controller coverage checks --------------------

	deviceController := GenericController[*models.Device, models.APIDevice, *models.APIDevice]{repo: repos.NewMockRepository[models.Device]()}
	deviceGroupController := GenericController[*models.DeviceGroup, models.APIDeviceGroup, *models.APIDeviceGroup]{repo: repos.NewMockRepository[models.DeviceGroup]()}
	orgController := GenericController[*models.Organization, models.APIOrg, *models.APIOrg]{repo: repos.NewMockRepository[models.Organization]()}
	envController := GenericController[*models.Environment, models.APIEnv, *models.APIEnv]{repo: repos.NewMockRepository[models.Environment]()}

	It("Incorrect Device URI on update", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		data, _ := json.Marshal(sampleDeviceBody(devicechangetype.DEV_GROUP_CHANGE.String()))
		context.Request, _ = http.NewRequest(http.MethodPut, "orgs/N86QADZN/devices/1337", bytes.NewBuffer(data))
		context.Params = []gin.Param{
			{Key: "deviceId", Value: "FOOBAR"},
			{Key: "orgKey", Value: orgKey},
		}
		deviceController.Update(context)
		Expect(context.Writer.Status()).To(BeEquivalentTo(http.StatusBadRequest))
	})
	It("Incorrect Device body on update", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest(http.MethodPut, "orgs/N86QADZN/devices/1337", bytes.NewBufferString("Bogus Data"))
		context.Params = []gin.Param{
			{Key: "deviceId", Value: "4567"},
			{Key: "orgKey", Value: orgKey},
		}
		deviceController.Update(context)
		Expect(context.Writer.Status()).To(BeEquivalentTo(http.StatusBadRequest))
	})
	It("Empty Device body on update", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest(http.MethodPut, "orgs/N86QADZN/devices/1337", bytes.NewBufferString("{}"))
		context.Params = []gin.Param{
			{Key: "deviceId", Value: "4567"},
			{Key: "orgKey", Value: orgKey},
		}
		deviceController.Update(context)
		Expect(context.Writer.Status()).To(BeEquivalentTo(http.StatusOK))
	})

	// Added for coverage

	It("Empty Device Group body on update", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest(http.MethodPut, "orgs/N86QADZN/devicegroups/1337", bytes.NewBufferString("{}"))
		context.Params = []gin.Param{
			{Key: "deviceGroupId", Value: "4567"},
			{Key: "orgKey", Value: orgKey},
		}
		deviceGroupController.Update(context)
		Expect(context.Writer.Status()).To(BeEquivalentTo(http.StatusOK))
	})

	It("Empty Org body on update", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest(http.MethodPut, "orgs/N86QADZN", bytes.NewBufferString("{}"))
		context.Params = []gin.Param{
			{Key: "orgKey", Value: orgKey},
		}
		orgController.Update(context)
		Expect(context.Writer.Status()).To(BeEquivalentTo(http.StatusOK))
	})

	It("Empty Env body on update", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest(http.MethodPut, "name/QA01", bytes.NewBufferString("{}"))
		context.Params = []gin.Param{
			{Key: "envName", Value: "QA01"},
		}
		envController.Update(context)
		Expect(context.Writer.Status()).To(BeEquivalentTo(http.StatusOK))
	})
})

func TestControllers(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "Controller Suite")
}
