// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package controllers

import (
	"bytes"
	"encoding/json"
	"net/http"
	"net/http/httptest"
	"testing"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/repos"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/deploymenttype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/enums/devicetype"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

const orgKey = "N86QADZN" // valid encoded OrgKey -- name="local", id=42
var now = time.Now().Round(1 * time.Second)

type BodyLogWriter struct {
	gin.ResponseWriter
	body *bytes.Buffer
}

func NewBodyLogWriter(writer gin.ResponseWriter) BodyLogWriter {
	return BodyLogWriter{ResponseWriter: writer, body: new(bytes.Buffer)}
}

func (w BodyLogWriter) Write(b []byte) (int, error) {
	w.body.Write(b)
	return w.ResponseWriter.Write(b)
}

func (w BodyLogWriter) WriteString(s string) (int, error) {
	w.body.WriteString(s)
	return w.ResponseWriter.WriteString(s)
}

func captureGinWriter(ctx *gin.Context) BodyLogWriter {
	blw := NewBodyLogWriter(ctx.Writer)
	ctx.Writer = blw
	return blw
}

func TestHintController_GetHints_NothingToDo(t *testing.T) {
	mockDeviceRepo := repos.NewMockRepository[models.Device]()
	mockGroupRepo := repos.NewMockRepository[models.DeviceGroup]()
	mockOrgRepo := repos.NewMockRepository[models.Organization]()
	mockGlobalRepo := repos.NewMockRepository[models.Environment]()

	thc := NewHintController(mockDeviceRepo, mockGroupRepo, mockOrgRepo, mockGlobalRepo)
	testContext, _ := gin.CreateTestContext(httptest.NewRecorder())

	hintReq := hints.NewHintRequestMessage(orgKey, 1337, 1337, 1, &now, &now, true, devicetype.WINDOWS, deploymenttype.ENDPOINT)
	data, _ := json.Marshal(hintReq)

	testContext.Request, _ = http.NewRequest(http.MethodPost, "/hints", bytes.NewBuffer(data))
	thc.GetHints(testContext)
	if testContext.Writer.Status() != http.StatusOK {
		t.Fatalf("Failed to process full device group adata")
	}
}

func TestHintController_GetHints_UpToDate(t *testing.T) {
	device := models.NewDeviceAtTime(123, orgKey, nil, nil, &now)
	group := models.NewDeviceGroup(1337, orgKey, &now, &now)
	org := models.NewOrganizationWithoutDeploymentTypeChanges(orgKey, true)
	env := models.NewEnvWithCanaryTimes("envName", &now, &now, now)

	mockDeviceRepo := repos.NewMockRepositoryWithModel(device)
	mockGroupRepo := repos.NewMockRepositoryWithModel(group)
	mockOrgRepo := repos.NewMockRepositoryWithModel(org)
	mockGlobalRepo := repos.NewMockRepositoryWithModel(env)

	thc := NewHintController(mockDeviceRepo, mockGroupRepo, mockOrgRepo, mockGlobalRepo)
	testContext, _ := gin.CreateTestContext(httptest.NewRecorder())

	hintReq := hints.NewHintRequestMessage(orgKey, 1337, 123, 1337, &now, &now, true, devicetype.WINDOWS, deploymenttype.ENDPOINT)
	data, _ := json.Marshal(hintReq)
	testContext.Request, _ = http.NewRequest(http.MethodPost, "/hints", bytes.NewBuffer(data))
	hintResponseBodyWriter := captureGinWriter(testContext)

	thc.GetHints(testContext)

	if testContext.Writer.Status() != http.StatusOK {
		t.Fatalf("Failed to process hint data")
	} else {
		hintResponseBody := hintResponseBodyWriter.body.Bytes()
		parsedHints := hints.NewHintResult()

		err := json.Unmarshal(hintResponseBody, &parsedHints)
		if err != nil {
			t.Fatalf("Failed to unmarshal: %v", err)
		}
		if parsedHints.AnyHintsSet() {
			t.Fatalf("Got hints when none were expected")
		}
	}
}

func TestHintController_GetHints_Hints(t *testing.T) {
	device := models.NewDeviceAtTime(123, orgKey, nil, nil, &now)
	group := models.NewDeviceGroup(1337, orgKey, &now, &now)
	org := models.NewOrganizationWithoutDeploymentTypeChanges(orgKey, true)
	env := models.NewEnv("envName", &now, &now)

	mockDeviceRepo := repos.NewMockRepositoryWithModel(device)
	mockGroupRepo := repos.NewMockRepositoryWithModel(group)
	mockOrgRepo := repos.NewMockRepositoryWithModel(org)
	mockGlobalRepo := repos.NewMockRepositoryWithModel(env)

	mockAuditor := MockAuditor{}
	thc := NewHintControllerWithAuditor(mockDeviceRepo, mockGroupRepo, mockOrgRepo, mockGlobalRepo,
		func(request *hints.HintRequestMessage, result *hints.HintResult) hints.Auditor {
			return &mockAuditor
		})
	testContext, _ := gin.CreateTestContext(httptest.NewRecorder())

	lastUpdateTime := time.Unix(0, 0)

	hintReq := hints.NewHintRequestMessage(orgKey, 1337, 1337, 1, &lastUpdateTime, &lastUpdateTime, false, devicetype.WINDOWS, deploymenttype.ENDPOINT)
	data, _ := json.Marshal(hintReq)
	testContext.Request, _ = http.NewRequest(http.MethodPost, "/hints", bytes.NewBuffer(data))
	hintResponseBodyWriter := captureGinWriter(testContext)

	thc.GetHints(testContext)

	if testContext.Writer.Status() != http.StatusOK {
		t.Fatalf("Failed to process hint data")
	} else {
		if mockAuditor.auditWasCalled {
			t.Fatalf("Audit was called, but should not have been")
		}

		hintResponseBody := hintResponseBodyWriter.body.Bytes()
		parsedHints := hints.NewHintResult()

		err := json.Unmarshal(hintResponseBody, &parsedHints)
		if err != nil {
			t.Fatalf("Failed to unmarshal: %v", err)
		}
		if !parsedHints.AnyHintsSet() {
			t.Fatalf("Did not get hints when hints were expected")
		}
	}
}

type MockAuditor struct {
	auditWasCalled bool
}

func (auditor *MockAuditor) Audit(ctx *gin.Context) {
	auditor.auditWasCalled = true
}

type spyHintGenerator struct {
	generator hints.IHintGenerator
	request   *hints.HintRequest
}

func (spy *spyHintGenerator) GenerateHints(request *hints.HintRequest, inAuditMode bool) *hints.HintResult {
	spy.request = request
	return spy.generator.GenerateHints(request, inAuditMode)
}

func TestHintController_GetHints_WithAuditHints(t *testing.T) {
	device := models.NewDeviceAtTime(123, orgKey, nil, nil, &now)
	group := models.NewDeviceGroup(1337, orgKey, &now, &now)
	org := models.NewOrganizationWithoutDeploymentTypeChanges(orgKey, true)
	env := models.NewEnv("envName", &now, &now)

	mockDeviceRepo := repos.NewMockRepositoryWithModel(device)
	mockGroupRepo := repos.NewMockRepositoryWithModel(group)
	mockOrgRepo := repos.NewMockRepositoryWithModel(org)
	mockGlobalRepo := repos.NewMockRepositoryWithModel(env)

	mockAuditor := MockAuditor{}
	thc := NewHintControllerWithAuditor(mockDeviceRepo, mockGroupRepo, mockOrgRepo, mockGlobalRepo,
		func(request *hints.HintRequestMessage, result *hints.HintResult) hints.Auditor {
			return &mockAuditor
		})
	testContext, _ := gin.CreateTestContext(httptest.NewRecorder())

	lastUpdateTime := time.Unix(0, 0)
	var auditLastUpdateTime int64 = 1674147375359
	auditHints := []string{"NEW_CONFIG_DATA", "NEW_POLICY", "ONETIME_HASH_DELETE_LIST"}
	auditHintsFromCBD := hints.NewAuditHintsFromCBD(auditHints, &auditLastUpdateTime, &auditLastUpdateTime)
	hintReq := hints.NewHintRequestMessageWithAuditHints(orgKey, 1337, 1337, 1,
		&lastUpdateTime, &lastUpdateTime, false, devicetype.WINDOWS, deploymenttype.ENDPOINT, auditHintsFromCBD)

	data, _ := json.Marshal(hintReq)
	testContext.Request, _ = http.NewRequest(http.MethodPost, "/hints", bytes.NewBuffer(data))
	hintResponseBodyWriter := captureGinWriter(testContext)

	thc.GetHints(testContext)
	if testContext.Writer.Status() != http.StatusOK {
		t.Fatalf("Failed to process hint data")
	} else {
		if !mockAuditor.auditWasCalled {
			t.Fatalf("Audit was not called")
		}

		hintResponseBody := hintResponseBodyWriter.body.Bytes()
		parsedHints := hints.NewHintResult()

		err := json.Unmarshal(hintResponseBody, &parsedHints)
		if err != nil {
			t.Fatalf("Failed to unmarshal: %v", err)
		}
		if !parsedHints.AnyHintsSet() {
			t.Fatalf("Did not get hints when hints were expected")
		}
	}
}

var _ = Describe("Hint Controller Coverage", func() {
	device := models.NewDeviceAtTime(123, orgKey, nil, nil, &now)
	group := models.NewDeviceGroup(1337, orgKey, &now, &now)
	org := models.NewOrganizationWithoutDeploymentTypeChanges(orgKey, true)
	env := models.NewEnv("envName", &now, &now)

	mockDeviceRepo := repos.NewMockRepositoryWithModel(device)
	mockGroupRepo := repos.NewMockRepositoryWithModel(group)
	mockOrgRepo := repos.NewMockRepositoryWithModel(org)
	mockGlobalRepo := repos.NewMockRepositoryWithModel(env)

	mockAuditor := MockAuditor{}
	thc := NewHintControllerWithAuditor(mockDeviceRepo, mockGroupRepo, mockOrgRepo, mockGlobalRepo,
		func(request *hints.HintRequestMessage, result *hints.HintResult) hints.Auditor {
			return &mockAuditor
		})

	lastUpdateTime := time.Unix(0, 0)

	// ----- Additional controller coverage checks --------------------

	It("Unparsable hinting body", func() {
		testContext, _ := gin.CreateTestContext(httptest.NewRecorder())
		testContext.Request, _ = http.NewRequest(http.MethodPost, "/hints", bytes.NewBufferString("Bogus"))

		thc.GetHints(testContext)
		Expect(testContext.Writer.Status()).To(BeEquivalentTo(http.StatusBadRequest))
	})
	It("Hinting body undecodable orgkey", func() {
		testContext, _ := gin.CreateTestContext(httptest.NewRecorder())
		hintReq := hints.NewHintRequestMessage("BOGUS", 1337, 1337, 1, &lastUpdateTime, &lastUpdateTime, false, devicetype.WINDOWS, deploymenttype.ENDPOINT)
		data, _ := json.Marshal(hintReq)
		testContext.Request, _ = http.NewRequest(http.MethodPost, "/hints", bytes.NewBuffer(data))

		thc.GetHints(testContext)
		Expect(testContext.Writer.Status()).To(BeEquivalentTo(http.StatusBadRequest))
	})

	It("Should get deviceType and deploymentType from HTTP json request body", func() {
		spy := &spyHintGenerator{generator: thc.generator}
		thc.withGenerator(spy)
		testContext, _ := gin.CreateTestContext(httptest.NewRecorder())
		hintReq := hints.NewHintRequestMessage("N86QADZN", 1337, 1337, 1, &lastUpdateTime, &lastUpdateTime, false, devicetype.WINDOWS, deploymenttype.ENDPOINT)
		hintReq.DeviceType = devicetype.WINDOWS
		hintReq.DeploymentType = deploymenttype.WORKLOAD
		data, _ := json.Marshal(hintReq)

		testContext.Request, _ = http.NewRequest(http.MethodPost, "/hints", bytes.NewBuffer(data))
		thc.GetHints(testContext)

		Expect(testContext.Writer.Status()).To(BeEquivalentTo(http.StatusOK))
		Expect(spy.request.Device.DeploymentType).ToNot(BeNil())
		Expect(*spy.request.Device.DeploymentType).To(BeComparableTo(deploymenttype.WORKLOAD))
		Expect(spy.request.Device.DeviceType).ToNot(BeNil())
		Expect(*spy.request.Device.DeviceType).To(BeComparableTo(devicetype.WINDOWS))
	})
})
