// Copyright (c) 2023-2024 VMware Carbon Black, all rights reserved

package middleware

import (
	"bytes"
	"context"
	"net/http"
	"net/http/httptest"
	"testing"

	"github.com/gin-gonic/gin"
	"gitlab.bit9.local/psc/platform/orgkey-go/v3"
	"gitlab.bit9.local/psckit/pschttp"
	"gitlab.bit9.local/psckit/pscrbac"
)

var (
	userOrgkeyStr = "K7MGPF37" // env: "unittest", orgId: 1
	userOrgkey, _ = orgkey.Decode(userOrgkeyStr)

	testRequestContext = pschttp.RequestContext{
		GatewayHost:           "example.com",
		GatewayPublicURL:      "example.com",
		GatewayPrivilegedPort: 1337,
		OriginEnv:             "unittest",
	}

	otherOrgkeyStr = "R76KWHRN" // env: "local", orgId: 1
)

type MockDemandService struct{}

func (mds *MockDemandService) HasPermission(_ context.Context, _ pschttp.Authentication, loginSession pschttp.UserOrgSession,
	targetOrgKey orgkey.OrgKey, permissions ...pscrbac.Permission,
) (bool, error) {
	// All operations other than "env" require an org key
	if permissions[0].Name != "csr.hints.env" {
		return targetOrgKey == loginSession.OrgKey, nil
	}

	// No org key is available for "env" operations
	return targetOrgKey.Env() == "", nil
}

func createUserContext(orgKey orgkey.OrgKey) pschttp.UserContext {
	return pschttp.UserContext{
		LoginSession: pschttp.UserOrgSession{
			OrgKey: *userOrgkey,
		},
		TargetSession: pschttp.UserOrgSession{
			OrgID:       1,
			UserID:      nil,
			UserName:    "test@example.com",
			OrgKey:      orgKey,
			ConnectorID: nil,
			RBACKey:     "psc:role:test",
		},
	}
}

func TestRBACMiddleWare_OK(t *testing.T) {
	// Assures that only users of the org can make changes to org devices
	c, _ := gin.CreateTestContext(httptest.NewRecorder())
	middleware := NewRBACMiddleware(&MockDemandService{}, pscrbac.NewPermission("csr.hints.device", pscrbac.Create))
	c.Set(pschttp.UserContextHeader, createUserContext(*userOrgkey))
	c.Set(pschttp.RequestContextHeader, testRequestContext)
	c.AddParam("org_key", userOrgkeyStr)
	c.Request = httptest.NewRequest("PUT", "/devices", bytes.NewBufferString("")).WithContext(context.TODO())
	middleware.getHandler()(c)
	if c.Writer.Status() != http.StatusOK {
		t.Errorf("Expected StatusOK, got %+v", c.Writer.Status())
	}
}

func TestRBACMiddleWare_NOTAUTHORIZED(t *testing.T) {
	// Assures that users of a different org are unable to make changes to another org's devices
	//  - NOTE: same behavior applies to devicegroups and orgs
	c, _ := gin.CreateTestContext(httptest.NewRecorder())
	middleware := NewRBACMiddleware(&MockDemandService{}, pscrbac.NewPermission("csr.hints.device", pscrbac.Create))

	c.Set(pschttp.UserContextHeader, createUserContext(*userOrgkey))
	c.Set(pschttp.RequestContextHeader, testRequestContext)

	// Use an org key that does NOT match the org associated with the login
	c.AddParam("org_key", otherOrgkeyStr)
	c.Request = httptest.NewRequest("PUT", "/devices", bytes.NewBufferString("")).WithContext(context.TODO())
	middleware.getHandler()(c)
	if c.Writer.Status() != http.StatusForbidden {
		t.Errorf("Expected StatusForbidden, got %+v", c.Writer.Status())
	}
}

func TestRBACMiddleWare_NoOrgKeyRequired(t *testing.T) {
	// Assures that environment changes are not restricted by user org, since env updates have no such information
	c, _ := gin.CreateTestContext(httptest.NewRecorder())
	middleware := NewRBACMiddleware(&MockDemandService{}, pscrbac.NewPermission("csr.hints.env", pscrbac.Create))

	c.Set(pschttp.UserContextHeader, createUserContext(*userOrgkey))
	c.Set(pschttp.RequestContextHeader, testRequestContext)
	c.Request = httptest.NewRequest("PUT", "/envs", bytes.NewBufferString("")).WithContext(context.TODO())
	middleware.getHandler()(c)
	if c.Writer.Status() != http.StatusOK {
		t.Errorf("Expected StatusOK, got %+v", c.Writer.Status())
	}
}

func TestRBACMiddleWare_NoOrgKeyWhenOneIsExpected(t *testing.T) {
	// Assures that missing orgkey is a problem when the user context has one specified
	//  - NOTE: same behavior applies to devicegroups and orgs
	c, _ := gin.CreateTestContext(httptest.NewRecorder())
	middleware := NewRBACMiddleware(&MockDemandService{}, pscrbac.NewPermission("csr.hints.device", pscrbac.Create))

	c.Set(pschttp.UserContextHeader, createUserContext(*userOrgkey))
	c.Set(pschttp.RequestContextHeader, testRequestContext)
	c.Request = httptest.NewRequest("PUT", "/devices", bytes.NewBufferString("")).WithContext(context.TODO())
	middleware.getHandler()(c)
	if c.Writer.Status() != http.StatusForbidden {
		t.Errorf("Expected StatusForbidden, got %+v", c.Writer.Status())
	}
}
