// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package middleware

import (
	"fmt"
	"net/http"
	"strings"
	"time"

	"github.com/google/uuid"
	"github.com/opentracing-contrib/go-stdlib/nethttp"
	"github.com/opentracing/opentracing-go"
	"github.com/opentracing/opentracing-go/ext"
	"gitlab.bit9.local/psc/ems/hinting/internal/logger"
	"go.uber.org/zap"
)

type contextKey string

var (
	ExtraLogFields  contextKey = "extraLogFields"
	AuditLogFields  contextKey = "auditLogFields"
	Status          contextKey = "status"
	Error           contextKey = "error"
	Errors          contextKey = "errors"
	SavedGinContext contextKey = "savedGinContext"
)

// ApplyMiddleware applies middleware to a handler
func (mwf RouteMiddlewareFactory) ApplyTracingMiddleware(hdlr http.Handler) http.Handler {
	if mwf.metricsClient != nil && mwf.metricsClient.Tracer() != nil {
		hdlr = nethttp.Middleware(
			mwf.metricsClient.Tracer(),
			hdlr,
			mwf.tracingOperationNameOption(),
			mwf.tracingSpanFilterOption(),
			mwf.tracingSpanObserverOption(),
		)
	}
	return hdlr
}

//goland:noinspection GoMixedReceiverTypes
func (mwf *RouteMiddlewareFactory) Logging(next http.Handler) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		defer func(start time.Time) {
			latency := time.Since(start)
			method := r.Method
			status := -1

			errorsExist := false

			ctx := r.Context()

			var savedGinContext SavedContext
			savedGinContextAny := ctx.Value(SavedGinContext)
			if savedGinContextAny != nil {
				if savedContext, ok := savedGinContextAny.(*SavedContext); ok {
					savedGinContext = *savedContext
				}
			}

			statusOrNil := ctx.Value(Status)
			if statusOrNil != nil {
				statusInt, ok := statusOrNil.(int)
				if ok {
					status = statusInt
				}
			}

			logFields := []zap.Field{
				zap.String("remote-address", r.RemoteAddr),
				zap.String("method", r.Method),
				zap.String("request-uri", r.RequestURI),
				zap.String("accept", strings.Join(r.Header["Accept"], " ")),
				zap.String("proto", r.Proto),
				zap.Int("status", status),
				zap.Int("length", int(r.ContentLength)),
				zap.String("user-agent", r.UserAgent()),
				zap.Duration("latency", latency),
			}

			shouldLog := false
			switch method {
			case "POST":
				shouldLog = false
				if savedGinContext.AuditLogFields != nil {
					shouldLog = true
					logFields = append(logFields, savedGinContext.AuditLogFields...)
				}
			case "GET":
				shouldLog = false
			case "PUT":
				shouldLog = true
			}

			extraLogFields := savedGinContext.ExtraLogFields
			if extraLogFields != nil {
				logFields = append(logFields, extraLogFields...)
			}

			errors := savedGinContext.Errs
			if errors != nil {
				logFields = append(logFields, zap.Strings("errors", errors))
				errorsExist = true
			}

			panicErrorHuh := ctx.Value(Error)
			if panicErrorHuh != nil {
				panicErr, ok := panicErrorHuh.(error)
				if ok {
					logFields = append(logFields, zap.NamedError("panic_err", panicErr))
					errorsExist = true
				}
			}

			switch status {
			case http.StatusInternalServerError:
				errorsExist = true
			case http.StatusServiceUnavailable:
				errorsExist = true
			case http.StatusRequestTimeout:
				errorsExist = true
			}

			switch {
			case errorsExist:
				logger.RequestLoggerFromContext(ctx).Error("", logFields...)
			case shouldLog:
				logger.RequestLoggerFromContext(ctx).Info("", logFields...)
			default:
				logger.RequestLoggerFromContext(ctx).Debug("", logFields...)
			}
		}(time.Now())

		next.ServeHTTP(w, r)
	}
}

//goland:noinspection GoMixedReceiverTypes
func (mwf RouteMiddlewareFactory) GetSpanErrorTaggingMiddleware(next http.HandlerFunc) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		defer func(w http.ResponseWriter, r *http.Request) {
			if sp := opentracing.SpanFromContext(r.Context()); sp != nil {
				shouldSetError := false
				errorInContext := r.Context().Value(Error)
				if errorInContext != nil {
					errValue, ok := errorInContext.(bool)
					shouldSetError = ok && errValue
				}
				setRetryAfter := func(w http.ResponseWriter) {
					w.Header().Set(RetryAfterHeaderKey, fmt.Sprintf("%d", mwf.cfg.RetryAfterSeconds))
				}
				statusOrNil := r.Context().Value(Status)
				if statusOrNil != nil && !shouldSetError {
					status, ok := statusOrNil.(int)
					if ok {
						switch status {
						case http.StatusInternalServerError:
							shouldSetError = true
						case http.StatusServiceUnavailable:
							shouldSetError = true
							setRetryAfter(w)
						case http.StatusRequestTimeout:
							shouldSetError = true
							setRetryAfter(w)
						default:
							shouldSetError = false
						}
					}
				}
				if shouldSetError {
					sp.SetTag(string(ext.Error), true)
				}
			}
		}(w, r)
		next(w, r)
	}
}

func (mwf RouteMiddlewareFactory) tracingOperationNameOption() nethttp.MWOption {
	return nethttp.OperationNameFunc(func(r *http.Request) string {
		splitPath := strings.Split(r.URL.Path, "/")
		method := r.Method
		var operationName string = "UKNOWN"
		switch method {
		case "PUT":
			operationName = strings.TrimSuffix(splitPath[len(splitPath)-2], "s")
		case "POST":
			operationName = splitPath[len(splitPath)-1]
		}
		return "HTTP " + r.Method + " " + operationName
	})
}

func (mwf RouteMiddlewareFactory) tracingSpanObserverOption() nethttp.MWOption {
	return nethttp.MWSpanObserver(func(sp opentracing.Span, r *http.Request) {
		sp.SetTag("http.remote_address", r.RemoteAddr)
		sp.SetTag("http.user_agent", r.UserAgent())
		sp.SetTag("http.accept", r.Header.Get("Accept"))
	})
}

// Tracing is a middleware to trace requests by id
func (mwf RouteMiddlewareFactory) TracingRequestIdAndLogger(next http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		var err error
		if sp := opentracing.SpanFromContext(r.Context()); sp != nil {
			err = opentracing.GlobalTracer().Inject(
				sp.Context(),
				opentracing.HTTPHeaders,
				opentracing.HTTPHeadersCarrier(w.Header()),
			)
		}
		requestID := w.Header().Get(mwf.cfg.RequestIDHeader)
		if requestID == "" {
			requestID = uuid.New().String()
			w.Header().Set(mwf.cfg.RequestIDHeader, requestID)
		}
		w.Header().Set(mwf.cfg.RequestIDHeader, requestID)
		l := mwf.RequestLogger.With(zap.String("request-id", requestID))
		if err != nil {
			l.Info(err.Error())
		}
		*r = *r.Clone(logger.ContextWithRequestLogger(r.Context(), l))
		next.ServeHTTP(w, r)
	})
}

func (mwf RouteMiddlewareFactory) tracingSpanFilterOption() nethttp.MWOption {
	return nethttp.MWSpanFilter(func(r *http.Request) bool {
		return !strings.Contains(r.RequestURI, "healthCheck")
	})
}
