// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package models

import (
	"bytes"
	"fmt"
	"net/http"
	"net/http/httptest"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	orgkeylib "gitlab.bit9.local/psc/platform/orgkey-go/v3"
)

var _ = Describe("API Environment validation tests", func() {
	orgKeyObject, _ := orgkeylib.Encode("env", 1337)
	envName := orgKeyObject.Env()

	// ----- URL Bind testing --------------------
	// NOTE: These handle supplied url API content validation, not value validation

	// NOTE: Same effect as supplying empty string
	It("Ensure we reject missing envName", func() {
		rawGroupString := `{"cert_service_time_changed":true, "policy_service_time_changed":true, "canary_payload_device_types":["WINDOWS", "IOS", "LINUX"]}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/env/uri", bytes.NewBufferString(rawGroupString))
		context.Params = []gin.Param{}
		apiEnv := APIEnv{}
		err := context.ShouldBindUri(apiEnv.GetURIPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIEnvURI.Name' Error:Field validation for 'Name' failed on the 'required' tag"))
	})

	// ----- Body Bind testing --------------------
	// NOTE: These handle supplied body API content validation, not value validation

	It("Ensure we can parse a valid env structure", func() {
		rawGroupString := `{"cert_service_time_changed":true, "policy_service_time_changed":true, "canary_payload_device_types":["WINDOWS", "IOS", "LINUX"]}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/env/uri", bytes.NewBufferString(rawGroupString))

		apiEnv := APIEnv{}
		err := context.ShouldBindJSON(apiEnv.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())

		Expect(apiEnv.PolicyServiceTimeChanged).To(BeEquivalentTo(true))
		Expect(apiEnv.CertServiceTimeChanged).To(BeEquivalentTo(true))
		Expect(len(apiEnv.CanaryPayloadDeviceTypes)).To(BeEquivalentTo(3))
	})

	It("Ensure we accept empty JSON body", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/env/uri", bytes.NewBufferString("{}"))

		apiEnv := APIEnv{}

		err := context.ShouldBindJSON(apiEnv.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(apiEnv.PolicyServiceTimeChanged).To(BeEquivalentTo(false))
		Expect(apiEnv.CertServiceTimeChanged).To(BeEquivalentTo(false))
		Expect(len(apiEnv.CanaryPayloadDeviceTypes)).To(BeEquivalentTo(0))
	})

	It("Ensure we reject an empty string body", func() {
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/env/uri", bytes.NewBufferString(""))

		apiEnv := APIEnv{}

		err := context.ShouldBindJSON(apiEnv.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("EOF")) // NOTE: best we can do with gomega
	})

	// -----

	// NOTE: Negative tests with booleans are limited due to how they are converted froma a string

	It("Ensure we accept only cert_service_time_changed", func() {
		rawGroupString := `{"cert_service_time_changed":true}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/env/uri", bytes.NewBufferString(rawGroupString))

		apiEnv := APIEnv{}
		err := context.ShouldBindJSON(apiEnv.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(apiEnv.CertServiceTimeChanged).To(BeEquivalentTo(true))
	})

	It("Ensure we accept only policy_service_time_changed", func() {
		rawGroupString := `{"policy_service_time_changed":true}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/env/uri", bytes.NewBufferString(rawGroupString))

		apiEnv := APIEnv{}
		err := context.ShouldBindJSON(apiEnv.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(apiEnv.PolicyServiceTimeChanged).To(BeEquivalentTo(true))
	})

	It("Ensure we reject non-list canary_payload_device_types", func() {
		rawGroupString := `{"canary_payload_device_types":"IOS"}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/env/uri", bytes.NewBufferString(rawGroupString))

		apiEnv := APIEnv{}
		err := context.ShouldBindJSON(apiEnv.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("cannot unmarshal string into Go struct field APIEnvBody.canary_payload_device_types of type []string"))
	})

	It("Ensure we allow empty canary_payload_device_types", func() {
		rawGroupString := `{"canary_payload_device_types":[]}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/env/uri", bytes.NewBufferString(rawGroupString))

		apiEnv := APIEnv{}
		err := context.ShouldBindJSON(apiEnv.GetBodyPart())
		Expect(err).ShouldNot(HaveOccurred())
		Expect(len(apiEnv.CanaryPayloadDeviceTypes)).To(BeEquivalentTo(0))
	})

	It("Ensure we reject excess canary_payload_device_types", func() {
		rawGroupString := `{"cert_service_time_changed":true, "canary_payload_device_types":["IOS", "WINDOWS", "ANDROID", "MAC", "LINUX", "OTHER", "MAC"]}`
		context, _ := gin.CreateTestContext(httptest.NewRecorder())
		context.Request, _ = http.NewRequest("PUT", "/some/env/uri", bytes.NewBufferString(rawGroupString))

		apiEnv := APIEnv{}
		err := context.ShouldBindJSON(apiEnv.GetBodyPart())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("'APIEnvBody.CanaryPayloadDeviceTypes' Error:Field validation for 'CanaryPayloadDeviceTypes' failed on the 'max' tag"))
	})

	// ----- DB Conversion testing --------------------
	// NOTE: These validate DB structure conversion checks

	useAPI := APIEnvURI{Name: envName}

	It("Ensure we convert a valid env structure", func() {
		apiEnv := APIEnv{
			APIEnvURI: useAPI,
			APIEnvBody: APIEnvBody{
				CertServiceTimeChanged: true, PolicyServiceTimeChanged: true,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
			},
		}

		env, err := apiEnv.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())

		Expect(env).ToNot(BeNil())
		Expect(env.CertServiceTime).ToNot(BeNil())
		Expect(env.PolicyServiceUpdateTime).ToNot(BeNil())
		Expect(len(env.CanaryPayloadTimestamps)).To(BeEquivalentTo(3))
	})

	It("Ensure we convert a valid env structure with false values", func() {
		apiEnv := APIEnv{
			APIEnvURI: useAPI,
			APIEnvBody: APIEnvBody{
				CertServiceTimeChanged: false, PolicyServiceTimeChanged: true,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
			},
		}

		env, err := apiEnv.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())

		Expect(env).ToNot(BeNil())
		Expect(env.CertServiceTime).To(BeNil())
		Expect(env.PolicyServiceUpdateTime).ToNot(BeNil())
		Expect(len(env.CanaryPayloadTimestamps)).To(BeEquivalentTo(3))
	})

	It("Ensure we reject Name as whitespace string", func() {
		apiEnv := APIEnv{
			APIEnvURI: APIEnvURI{Name: "      "},
			APIEnvBody: APIEnvBody{
				CertServiceTimeChanged: true, PolicyServiceTimeChanged: true,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
			},
		}

		env, err := apiEnv.ConvertToDBObject()
		Expect(env).To(BeNil())
		Expect(err).Should(HaveOccurred())
		Expect(fmt.Sprintln(err)).To(ContainSubstring("failed to convert Env.Name (cannot be empty string or all whitespace)"))
	})

	It("Ensure we accept Name with trailing/leading whitespace but trim it (and give warning)", func() {
		apiEnv := APIEnv{
			APIEnvURI: APIEnvURI{Name: "\tenv1234\n"},
			APIEnvBody: APIEnvBody{
				CertServiceTimeChanged: true, PolicyServiceTimeChanged: true,
				CanaryPayloadDeviceTypes: []string{"WINDOWS", "LINUX", "IOS"},
			},
		}

		env, err := apiEnv.ConvertToDBObject()
		Expect(err).ShouldNot(HaveOccurred())

		Expect(env).ToNot(BeNil())
		Expect(env.Name).To(BeEquivalentTo("env1234"))
	})

	It("Ensure we reject an invalid DeviceType for CanaryPayloadDeviceTypes", func() {
		apiEnv := APIEnv{
			APIEnvURI: useAPI,
			APIEnvBody: APIEnvBody{
				CertServiceTimeChanged: true, PolicyServiceTimeChanged: true,
				CanaryPayloadDeviceTypes: []string{"BOGUS_DEVICE_TYPE"},
			},
		}

		env, err := apiEnv.ConvertToDBObject()
		Expect(env).ToNot(BeNil())
		Expect(err).Should(MatchError(ErrModelConversionWarning))
	})

	// NOTE: Duplicate CanaryPayloadDeviceTypes tested in functional tests (requires live service)
})
