// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

// Package models -- contains database models
package models

import (
	"fmt"
	"time"

	"gorm.io/gorm"
	"gorm.io/gorm/clause"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/devicegroupchangetype"
	"gitlab.bit9.local/psc/ems/hinting/internal/utils"

	"go.uber.org/zap/zapcore"
)

////////////////////////////////////////////////////////////////////////////////
// Postgres Database Specifics

// DeviceGroup -- device group structure stored in postgres
type DeviceGroup struct {
	Id                   int64  `gorm:"primaryKey"`
	OrgKey               string `gorm:"primaryKey"`
	env                  string
	LiveQueryChangeTime  *time.Time
	LastPolicyChangeTime *time.Time
}

func (deviceGroup *DeviceGroup) CacheKey() string {
	return fmt.Sprintf("device_groups-{%s}-%d", deviceGroup.getOrgKey(), deviceGroup.Id)
}

func (deviceGroup *DeviceGroup) getOrgKey() string {
	return deviceGroup.OrgKey
}

func (deviceGroup *DeviceGroup) GetEnv() string {
	return deviceGroup.env
}

func (deviceGroup *DeviceGroup) GetPrimaryColumnNames() []clause.Column {
	return []clause.Column{{Name: "id"}, {Name: "org_key"}}
}

func (deviceGroup *DeviceGroup) GetOnConflictClause() clause.OnConflict {
	return clause.OnConflict{
		Columns:   deviceGroup.GetPrimaryColumnNames(),
		DoUpdates: deviceGroup.GetUpsertClauses(),
	}
}

func (deviceGroup *DeviceGroup) GetUpsertClauses() clause.Set {
	return clause.Assignments(map[string]interface{}{
		"live_query_change_time":  gorm.Expr("GREATEST(device_groups.live_query_change_time, excluded.live_query_change_time)"),
		"last_policy_change_time": gorm.Expr("GREATEST(device_groups.last_policy_change_time, excluded.last_policy_change_time)"),
	})
}

func (deviceGroup *DeviceGroup) MarshalLogObject(encoder zapcore.ObjectEncoder) error {
	encoder.AddInt64("group_id", deviceGroup.Id)
	encoder.AddString("org_key", deviceGroup.OrgKey)
	encoder.AddString("env", deviceGroup.env)
	utils.AddTimePointerToEncoder(encoder, "live_query_change_time", deviceGroup.LiveQueryChangeTime)
	utils.AddTimePointerToEncoder(encoder, "last_policy_change_time", deviceGroup.LastPolicyChangeTime)
	return nil
}

//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// API Input Handling

type APIDeviceGroupURI struct {
	Id     int64  `uri:"deviceGroupId" binding:"required"`
	OrgKey string `uri:"orgKey" binding:"required"`
}

type APIDeviceGroupBody struct {
	// List of device group change types to be updated.
	//
	// Valid Change Types:
	// DG_LIVE_QUERY_CHANGE,
	// DG_POLICY_CHANGE
	//
	// required: false
	// example: ["DG_POLICY_CHANGE", "DG_LIVE_QUERY_CHANGE"]
	Changes []string `json:"changes" binding:"omitempty,max=2"`

	// If specified, overrides the current time value for any DG_LIVE_QUERY_CHANGE.  Has no effect with other change types.
	//
	// required: false
	// example: 1672531200000
	LiveQueryTime *int64 `json:"live_query_time"`
}

type APIDeviceGroup struct {
	APIDeviceGroupURI
	APIDeviceGroupBody
}

// GetBodyPart -- Get the APIDeviceGroupBody component
func (apiDeviceGroup *APIDeviceGroup) GetBodyPart() any {
	return &apiDeviceGroup.APIDeviceGroupBody
}

// GetURIPart -- Get the APIDeviceGroupURI component
func (apiDeviceGroup *APIDeviceGroup) GetURIPart() URIPart[*DeviceGroup] {
	return &apiDeviceGroup.APIDeviceGroupURI
}

// ConvertToDBObject -- Initialize a DeviceGroup with a received APIDeviceGroupURI
func (uri *APIDeviceGroupURI) ConvertToDBObject() (*DeviceGroup, error) {
	uri.OrgKey = utils.Sanitize("orgkey", uri.OrgKey)
	_, env, err := utils.ValidateOrgKey(uri.OrgKey)
	if err != nil {
		return nil, err
	}
	return &DeviceGroup{Id: uri.Id, OrgKey: uri.OrgKey, env: env}, nil
}

// ConvertToDBObject -- Translates the APIDeviceGroup structure obtained from a user/other service into a DeviceGroup structure
// that is stored in the postgres database.
func (apiDeviceGroup *APIDeviceGroup) ConvertToDBObject() (deviceGroup *DeviceGroup, err error) {
	now := time.Now()
	deviceGroup, err = apiDeviceGroup.APIDeviceGroupURI.ConvertToDBObject()
	if err != nil {
		return nil, err
	}

	// NOTE: Additions/removals here will require updates in `test/functional/suites/lib/models/enumerations.py`
	//       as well as changes in `internal/app/models/enums/devicegroupchangetype/devicegroupchangetype.go`
	for _, change := range apiDeviceGroup.Changes {
		switch change {
		case devicegroupchangetype.DG_LIVE_QUERY_CHANGE.String():
			if apiDeviceGroup.LiveQueryTime == nil {
				err = NewModelConversionWarning("Device Group DG_LIVE_QUERY_CHANGE did NOT receive live_query_time")
				deviceGroup.LiveQueryChangeTime = &now
			} else {
				deviceGroup.LiveQueryChangeTime = utils.GetTimePointer(*apiDeviceGroup.LiveQueryTime)
			}
		case devicegroupchangetype.DG_POLICY_CHANGE.String():
			deviceGroup.LastPolicyChangeTime = &now
		default:
			err = NewModelConversionWarning("Unknown DeviceGroupChangeType '%s'", change)
		}
	}

	return deviceGroup, err
}

//
////////////////////////////////////////////////////////////////////////////////

// NewDeviceGroup -- Create a usable DeviceGroup for testing purposes
func NewDeviceGroup(id int64, orgKey string, liveQueryChangeTime, changeTime *time.Time) *DeviceGroup {
	return &DeviceGroup{Id: id, OrgKey: orgKey, LiveQueryChangeTime: liveQueryChangeTime, LastPolicyChangeTime: changeTime}
}
