// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

// Package orgchangetype -- Contains enumeration values for organization change Types
package orgchangetype

//go:generate stringer -type=OrgChangeType

// OrgChangeType -- Enum of allowed organization change types
type OrgChangeType int

// NOTE: Changes here should also be reflected in `test/functional/suites/lib/models/enumerations.py`

// NOTE: Keep in sync with:
//   - https://gitlab.bit9.local/cb-defense/cbd/-/blob/main/backend/domainModel/src/main/java/com/scargo/domainmodel/entity/support/ChangeTimestampType.java
//

//goland:noinspection GoSnakeCaseUsage
const (
	ORG_NET_CFG_CHANGE   OrgChangeType = 101 // Network config for the org has changed
	ORG_HASH_LIST_CHANGE OrgChangeType = 102 // Org black list/white list/gray list has changed

	// NOTE: The following change types not used for hinting, but we are aware of them:
	//  - ORG_YARA_AV                   (103)
	//  - ORG_YARA_AV                   (103)
	//  - ORG_YARA_IP                   (104)
	//  - ORG_YARA_PE_HEADER            (105)
	//  - ORG_SENSOR_TESTKIT_UPDATE     (106)
	//  - ORG_DISABLE_SENSOR_AUTOUPDATE (107)

	ORG_AV_INTEGRATION_CHANGE     OrgChangeType = 108 // Org AV integration change
	ORG_KILL_SWITCH_ON            OrgChangeType = 109 // Org level kill switch on for CbLR
	ORG_UNINSTALL_CODE_CHANGE     OrgChangeType = 110 // Org level uninstall code changed
	ORG_LIVE_QUERY_CHANGE         OrgChangeType = 111 // Org level live query
	ORG_SET_MELTDOWN_REGISTRY_KEY OrgChangeType = 112 // Set meltdown registry key timestamp
	ORG_THREATHUNTER_RULES        OrgChangeType = 113 // Threat hunter rules changed
	ORG_DISABLE_DEFENSE_RULES     OrgChangeType = 114 // Org Disable Defense rules hint
	ORG_BLADE_ACTION_CHANGE       OrgChangeType = 115 // Request to trigger blade actions
)

//goland:noinspection GoSnakeCaseUsage
var (
	// AllOrgChangeTypes -- List of OrgChangeType that is used by hinting
	//
	AllOrgChangeTypes = []OrgChangeType{ //nolint:gochecknoglobals // we don't care
		ORG_NET_CFG_CHANGE,
		ORG_HASH_LIST_CHANGE,
		ORG_AV_INTEGRATION_CHANGE,
		ORG_KILL_SWITCH_ON,
		ORG_UNINSTALL_CODE_CHANGE,
		ORG_LIVE_QUERY_CHANGE,
		ORG_SET_MELTDOWN_REGISTRY_KEY,
		ORG_THREATHUNTER_RULES,
		ORG_DISABLE_DEFENSE_RULES,
		ORG_BLADE_ACTION_CHANGE,
	}

	// TranslateOrgChangeTypeString -- Translation from string -> OrgChangeType
	TranslateOrgChangeTypeString = map[string]OrgChangeType{ //nolint:gochecknoglobals // we don't care
		ORG_NET_CFG_CHANGE.String():            ORG_NET_CFG_CHANGE,
		ORG_HASH_LIST_CHANGE.String():          ORG_HASH_LIST_CHANGE,
		ORG_AV_INTEGRATION_CHANGE.String():     ORG_AV_INTEGRATION_CHANGE,
		ORG_KILL_SWITCH_ON.String():            ORG_KILL_SWITCH_ON,
		ORG_UNINSTALL_CODE_CHANGE.String():     ORG_UNINSTALL_CODE_CHANGE,
		ORG_LIVE_QUERY_CHANGE.String():         ORG_LIVE_QUERY_CHANGE,
		ORG_SET_MELTDOWN_REGISTRY_KEY.String(): ORG_SET_MELTDOWN_REGISTRY_KEY,
		ORG_THREATHUNTER_RULES.String():        ORG_THREATHUNTER_RULES,
		ORG_DISABLE_DEFENSE_RULES.String():     ORG_DISABLE_DEFENSE_RULES,
		ORG_BLADE_ACTION_CHANGE.String():       ORG_BLADE_ACTION_CHANGE,
	}
)
