// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package types

import (
	"database/sql/driver"
	"encoding/json"
	"errors"
	"fmt"
	"time"

	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"

	"gitlab.bit9.local/psc/ems/hinting/internal/logger"
	"gitlab.bit9.local/psc/ems/hinting/internal/utils"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models/enums/orgchangetype"
)

// OrgChangeset -- A structure used by the Org model
type OrgChangeset map[orgchangetype.OrgChangeType]*time.Time

//goland:noinspection GoMixedReceiverTypes
func (orgChanges OrgChangeset) MarshalLogObject(encoder zapcore.ObjectEncoder) error {
	for changeType, timePointer := range orgChanges {
		utils.AddTimePointerToEncoder(encoder, changeType.String(), timePointer)
	}
	return nil
}

// ----- Processing methods for OrgChangeset --------------------

// NewEmptyChangeset -- Create an empty OrgChangeset for later use
func NewEmptyChangeset() OrgChangeset {
	return make(OrgChangeset)
}

// NewChangesetAtTime -- Create a map of all OrgChangeTypes with the supplied time
func NewChangesetAtTime(t time.Time) OrgChangeset {
	times := NewEmptyChangeset()
	for _, changeType := range orgchangetype.AllOrgChangeTypes {
		times[changeType] = &t
	}
	return times
}

// GetTimeForOrgChange -- Get the time for a given OrgChangeType, or nil if not defined yet
//
//goland:noinspection GoMixedReceiverTypes
func (orgChanges OrgChangeset) GetTimeForOrgChange(changeType orgchangetype.OrgChangeType) *time.Time {
	timestamp, found := orgChanges[changeType]
	if found {
		return timestamp
	} else {
		return nil
	}
}

// NewOrgChangeSet -- generate a new organization change set based on a supplied map (The inverse of toMap)
//   - Converts a 'map[string]millis' structure into a 'map[OrgChangeType]timestamp' structure
func NewOrgChangeSet(changes map[string]int64, source zap.Field) OrgChangeset {
	oc := make(OrgChangeset)
	for change := range changes {
		changeType, found := orgchangetype.TranslateOrgChangeTypeString[change]
		if found {
			timeLong, found := changes[change]
			if found {
				convertedTime := time.UnixMilli(timeLong)
				oc[changeType] = &convertedTime
			}
		} else {
			logger.BaseLogger().Warn(fmt.Sprintf("Unknown OrgChangeType '%s' seen", change), source)
		}
	}
	return oc
}

// Scan -- Convert incoming bytes into a OrgChangeset structure
//
//goland:noinspection GoMixedReceiverTypes
func (orgChanges *OrgChangeset) Scan(value interface{}) error {
	bytes, ok := value.([]byte)
	if !ok {
		return errors.New(fmt.Sprint("Failed to unmarshal JSONB value:", value))
	}
	return json.Unmarshal(bytes, orgChanges)
}

// Value -- return json value of OrgChangeset, implement driver.Value interface
//
//goland:noinspection GoMixedReceiverTypes
func (orgChanges OrgChangeset) Value() (driver.Value, error) {
	if len(orgChanges) == 0 {
		return nil, nil
	}
	rawBytes, err := json.Marshal(orgChanges)
	if err != nil {
		return nil, err
	}
	return json.RawMessage(rawBytes).MarshalJSON()
}
