// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package repos

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"

	"github.com/opentracing/opentracing-go"
	"gitlab.bit9.local/psc/ems/hinting/internal/app/models"
	"gitlab.bit9.local/psc/ems/hinting/internal/dbwrapper"
	"gitlab.bit9.local/psc/ems/hinting/internal/redisclient"
)

type CachedRepository[T models.Model] struct {
	redisClient redisclient.RedisClient
	repo        Repository[T]
}

func NewCachedRepoFromRepo[T models.Model](redisClient redisclient.RedisClient, repo Repository[T]) CachedRepository[T] {
	return CachedRepository[T]{redisClient: redisClient, repo: repo}
}

func NewCachedRepo[T models.Model](redisClient redisclient.RedisClient, db *dbwrapper.DbClient) CachedRepository[T] {
	return CachedRepository[T]{redisClient: redisClient, repo: GenericRepository[T]{db: db}}
}

func (c CachedRepository[T]) Get(ctx context.Context, model T) (err error) {
	if sp := opentracing.SpanFromContext(ctx); sp != nil {
		childSp := opentracing.StartSpan("CACHE GET", opentracing.ChildOf(sp.Context()))
		defer childSp.Finish()
	}
	return c.handleCacheGet(ctx, model)
}

func (c CachedRepository[T]) handleCacheGet(ctx context.Context, model T) (err error) {
	bytes, err := c.redisClient.Get(ctx, model.CacheKey())

	switch {
	case err == nil:
		if json.Unmarshal([]byte(bytes), model) != nil {
			// Something was wrong with the cached data so let's pretend it wasn't there and "refresh" it
			err = c.handleCacheMiss(ctx, model)
		}
		return
	case errors.Is(err, context.DeadlineExceeded):
		return fmt.Errorf("redis timeout during Get: %w", err)
	case errors.Is(err, redisclient.ErrRedisNil):
		return c.handleCacheMiss(ctx, model)
	default:
		return
	}
}

func (c CachedRepository[T]) handleCacheMiss(ctx context.Context, model T) (err error) {
	err = c.repo.Get(ctx, model)
	if err == nil {
		return c.handleCacheSet(ctx, model)
	}

	return err
}

func (c CachedRepository[T]) handleCacheSet(ctx context.Context, model T) (err error) {
	bytes, err := json.Marshal(model)
	if err == nil {
		sp := opentracing.SpanFromContext(ctx)
		if sp != nil {
			childSp := opentracing.StartSpan("CACHE SET", opentracing.ChildOf(sp.Context()))
			defer childSp.Finish()
		}

		err = c.redisClient.Set(ctx, model.CacheKey(), bytes)
		if errors.Is(err, context.DeadlineExceeded) {
			return fmt.Errorf("redis timeout during Set: %w", err)
		}
	}
	return err
}

func (c CachedRepository[T]) Update(ctx context.Context, model T) (err error) {
	err = c.repo.Update(ctx, model)
	if err == nil {
		err = c.handleCacheSet(ctx, model)
	}
	return
}
