// Copyright (c) 2022-2023 VMware Carbon Black, all rights reserved

package repos

import (
	"context"
	"encoding/json"
	"fmt"
	"testing"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/internal/app/models"
	"gitlab.bit9.local/psc/ems/hinting/internal/redisclient"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

func NewDeviceMatcher(device *models.Device) *DeviceMatcher {
	return &DeviceMatcher{Device: device}
}

type DeviceMatcher struct {
	*models.Device
	err error
}

func timesAboutEqual(expected, actual *time.Time) bool {
	if expected == nil {
		if actual == nil {
			return true
		} else {
			return false
		}
	} else {
		if actual == nil {
			return false
		}
		roundedExpected := expected.Round(100 * time.Millisecond)
		roundedActual := actual.Round(100 * time.Millisecond)
		isEqual := roundedExpected.Equal(roundedActual)
		if !isEqual {
			fmt.Printf("Timestamps mismatch. Actual: %s vs Expected: %s\n", roundedActual, roundedExpected)
		}
		return isEqual
	}
}

// Match -- This variant compares two devices for equality
func (deviceMatcher *DeviceMatcher) Match(actual interface{}) (success bool, err error) {
	device := actual.(models.Device)
	if deviceMatcher.Id != device.Id {
		deviceMatcher.err = fmt.Errorf("ids didn't match")
		return false, deviceMatcher.err
	}
	if deviceMatcher.OrgKey != device.OrgKey {
		deviceMatcher.err = fmt.Errorf("'OrgKey' didn't match")
		return false, deviceMatcher.err
	}

	if !timesAboutEqual(deviceMatcher.GroupChangeTime, device.GroupChangeTime) {
		deviceMatcher.err = fmt.Errorf("'GroupChangeTime' didn't match")
		return false, deviceMatcher.err
	}

	if !timesAboutEqual(deviceMatcher.SensorConfigUpdateTime, device.SensorConfigUpdateTime) {
		deviceMatcher.err = fmt.Errorf("'SensorConfigUpdateTime' didn't match")
		return false, deviceMatcher.err
	}

	if !timesAboutEqual(deviceMatcher.SensorCertificateUpdateTime, device.SensorCertificateUpdateTime) {
		deviceMatcher.err = fmt.Errorf("'SensorCertificateUpdateTime' didn't match")
		return false, deviceMatcher.err
	}

	if !timesAboutEqual(deviceMatcher.HashDeleteRequestTime, device.HashDeleteRequestTime) {
		deviceMatcher.err = fmt.Errorf("'HashDeleteRequestTime' didn't match")
		return false, deviceMatcher.err
	}

	if !timesAboutEqual(deviceMatcher.HashUploadRequestTime, device.HashUploadRequestTime) {
		deviceMatcher.err = fmt.Errorf("'HashUploadRequestTime' didn't match")
		return false, deviceMatcher.err
	}

	if !timesAboutEqual(deviceMatcher.LogsRequestTime, device.LogsRequestTime) {
		deviceMatcher.err = fmt.Errorf("'LogsRequestTime' didn't match")
		return false, deviceMatcher.err
	}

	if !timesAboutEqual(deviceMatcher.SensorKitUpdateTime, device.SensorKitUpdateTime) {
		deviceMatcher.err = fmt.Errorf("'SensorKitUpdateTime' didn't match")
		return false, deviceMatcher.err
	}

	if !timesAboutEqual(deviceMatcher.PerfmonTriggerTime, device.PerfmonTriggerTime) {
		deviceMatcher.err = fmt.Errorf("PerfmonTriggerTime didn't match")
		return false, deviceMatcher.err
	}

	if !timesAboutEqual(deviceMatcher.SensorActionTime, device.SensorActionTime) {
		deviceMatcher.err = fmt.Errorf("'PerfmonTriggerTime' didn't match")
		return false, deviceMatcher.err
	}

	if !timesAboutEqual(deviceMatcher.ReRegisterTime, device.ReRegisterTime) {
		deviceMatcher.err = fmt.Errorf("'ReRegisterTime' didn't match")
		return false, deviceMatcher.err
	}

	if !timesAboutEqual(deviceMatcher.LiveQueryChangeTime, device.LiveQueryChangeTime) {
		deviceMatcher.err = fmt.Errorf("'LiveQueryChangeTime' didn't match")
		return false, deviceMatcher.err
	}

	if deviceMatcher.DeploymentType == nil {
		if device.DeploymentType != nil {
			deviceMatcher.err = fmt.Errorf("'DeploymentType' didn't match; nil vs. %s", *device.DeploymentType)
			return false, deviceMatcher.err
		}
	} else {
		if device.DeploymentType == nil {
			deviceMatcher.err = fmt.Errorf("'DeploymentType' didn't match; %s vs. nil", *deviceMatcher.DeploymentType)
			return false, deviceMatcher.err
		} else if *(device.DeploymentType) != *(deviceMatcher.DeploymentType) {
			deviceMatcher.err = fmt.Errorf("DeploymentType didn't match; %s vs. %s", *deviceMatcher.DeploymentType, *device.DeploymentType)
			return false, deviceMatcher.err
		}
	}

	if device.DeviceType != deviceMatcher.DeviceType {
		deviceMatcher.err = fmt.Errorf("'DeviceTypeId' didn't match; %s vs. %s", *deviceMatcher.DeviceType, *device.DeviceType)
		return false, deviceMatcher.err
	}

	return true, nil
}

func (deviceMatcher *DeviceMatcher) FailureMessage(_ interface{}) (message string) {
	return fmt.Sprintf("Failed: %v", deviceMatcher.err)
}

func (deviceMatcher *DeviceMatcher) NegatedFailureMessage(_ interface{}) (message string) {
	return fmt.Sprintf("Negated Failure: %v", deviceMatcher.err)
}

type mockRedisClient struct {
	redisclient.RedisClient
	getter func(ctx context.Context, cacheKey string) ([]byte, error)
	setter func(ctx context.Context, cacheKey string, data []byte) error
	deler  func(ctx context.Context, cacheKey string) error
}

func (mrc *mockRedisClient) Get(ctx context.Context, cacheKey string) ([]byte, error) {
	return mrc.getter(ctx, cacheKey)
}

func (mrc *mockRedisClient) Set(ctx context.Context, cacheKey string, data []byte) error {
	return mrc.setter(ctx, cacheKey, data)
}

func (mrc *mockRedisClient) Del(ctx context.Context, cacheKey string) error {
	return mrc.deler(ctx, cacheKey)
}

const cacheID int64 = 777

// ----- Begin Tests --------------------

var _ = Describe("The cached repository", func() {
	dbDevice := models.NewDeviceAtTime(777, "orgkey", nil, nil, &now)
	// dbBytes, _ := json.Marshal(dbDevice)
	cacheDevice := models.NewDeviceAtTime(cacheID, "orgkey", nil, nil, &now)
	cacheBytes, marshalError := json.Marshal(cacheDevice)
	Expect(marshalError).To(BeNil())
	underlyingRepoWithModel := NewMockRepositoryWithModel(dbDevice)

	underlyingRepoWithoutModel := NewMockRepositoryWithError[models.Device]()

	Context("Retrieval tests", func() {
		It("An item in the cache is returned", func() {
			cachedRepoWithModel := NewCachedRepoFromRepo(&mockRedisClient{
				getter: func(ctx context.Context, cacheKey string) ([]byte, error) {
					return cacheBytes, nil
				},
			}, underlyingRepoWithModel)
			device := models.Device{Id: cacheDevice.Id, OrgKey: cacheDevice.OrgKey}
			err := cachedRepoWithModel.Get(context.TODO(), &device)
			Expect(err).To(BeNil())
			Expect(device).To(NewDeviceMatcher(cacheDevice))
		})

		It("Falls back to Database if cache is empty", func() {
			cachedRepoWithModel := NewCachedRepoFromRepo(&mockRedisClient{
				getter: func(ctx context.Context, cacheKey string) ([]byte, error) {
					return nil, redisclient.ErrRedisNil
				},
				setter: func(ctx context.Context, cacheKey string, data []byte) error {
					return nil
				},
			}, underlyingRepoWithModel)
			device := models.Device{Id: dbDevice.Id, OrgKey: dbDevice.OrgKey}
			err := cachedRepoWithModel.Get(context.TODO(), &device)
			Expect(err).To(BeNil())
			Expect(device).To(NewDeviceMatcher(dbDevice))
		})

		It("Error is reported when data is not found", func() {
			cachedRepoWithoutModel := NewCachedRepoFromRepo(&mockRedisClient{
				getter: func(ctx context.Context, cacheKey string) ([]byte, error) {
					return nil, redisclient.ErrRedisNil
				},
			}, underlyingRepoWithoutModel)
			device := models.Device{Id: cacheDevice.Id, OrgKey: cacheDevice.OrgKey}
			err := cachedRepoWithoutModel.Get(context.TODO(), &device)
			Expect(err).ToNot(BeNil())
			Expect(fmt.Sprintln(err)).To(ContainSubstring("a database error has occurred"))
		})
	})

	Context("Timeout tests", func() {
		It("Error is reported when Get timeout occurs", func() {
			cachedRepoWithModel := NewCachedRepoFromRepo(&mockRedisClient{
				getter: func(context.Context, string) ([]byte, error) {
					return nil, context.DeadlineExceeded
				},
			}, underlyingRepoWithModel)
			device := models.Device{Id: cacheDevice.Id, OrgKey: cacheDevice.OrgKey}
			err := cachedRepoWithModel.Get(context.Background(), &device)
			Expect(err).ToNot(BeNil())
			Expect(fmt.Sprintln(err)).To(ContainSubstring("context deadline exceeded"))
		})
	})

	Context("Update tests", func() {
		It("Item is updated in database and redis", func() {
			redisSetCalled := false
			cachedRepoWithModel := NewCachedRepoFromRepo(&mockRedisClient{
				setter: func(ctx context.Context, cacheKey string, data []byte) error {
					redisSetCalled = true
					return nil
				},
			}, underlyingRepoWithModel)
			err := cachedRepoWithModel.Update(context.TODO(), dbDevice)
			Expect(err).To(BeNil())
			Expect(redisSetCalled).To(BeTrue())
		})
		It("errors are returned when update fails", func() {
			redisSetCalled := false
			cachedRepoWithModel := NewCachedRepoFromRepo(&mockRedisClient{
				setter: func(ctx context.Context, cacheKey string, data []byte) error {
					redisSetCalled = true
					return fmt.Errorf("update error")
				},
			}, underlyingRepoWithModel)
			err := cachedRepoWithModel.Update(context.TODO(), dbDevice)
			Expect(err).ToNot(BeNil())
			Expect(redisSetCalled).To(BeTrue())
		})
	})
})

func TestCacheRepo(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "Cached Repo Suite")
}
