// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package config

import (
	"flag"
	"os"
	"sync"
	"time"

	"github.com/spf13/viper"
	"gorm.io/gorm"
)

// singleton configuration instance
var (
	onceConfig   sync.Once
	instance     *Config
	profilerFlag *bool
)

func ParseCliFlags() {
	profilerFlag = flag.Bool("profiler", false, "Enable profiling")
	flag.Parse()
}

func CreateConfigFromFile(path, filename, configType string) (cfg *Config, err error) {
	viper.AddConfigPath(path)
	viper.SetConfigName(filename)
	viper.SetConfigType(configType)
	err = viper.ReadInConfig()
	if err == nil {
		cfg, err = newConfigFromEnv(false)
	}
	return
}

func reset() {
	cfg, err := newConfigFromEnv(true)
	if err != nil {
		panic("Failed to load config from environment")
	}
	instance = cfg
}

func GetInstance() *Config {
	onceConfig.Do(func() {
		reset()
	})
	return instance
}

type IConfigParam[T any] interface {
	bind()
	establishDefault()
	Get(bind bool) T
}

type GenericEnvConfigParam[T any] struct {
	envar        string
	defaultValue T
	hasDefault   bool
}
type StringParam struct {
	GenericEnvConfigParam[string]
}

func (sp StringParam) Get(bind bool) string {
	if bind {
		sp.bind()
	}
	return viper.GetString(sp.envar)
}

type DurationParam struct {
	GenericEnvConfigParam[time.Duration]
	unit time.Duration
}

func (dp DurationParam) Get(bind bool) time.Duration {
	if bind {
		dp.bind()
	}
	return viper.GetDuration(dp.envar) * dp.unit
}

func NewConfigParam(envar string) IConfigParam[string] {
	return StringParam{GenericEnvConfigParam: NewGenericConfigParamWithoutDefault(envar)}
}

func NewConfigParamWithDefault(envar, defaultValue string) IConfigParam[string] {
	return StringParam{GenericEnvConfigParam: NewGenericConfigParam(envar, defaultValue)}
}

func NewDurationConfigParam(envar string, defaultValue, unit time.Duration) IConfigParam[time.Duration] {
	return DurationParam{GenericEnvConfigParam: NewGenericConfigParam(envar, defaultValue), unit: unit}
}

type IntParam struct {
	GenericEnvConfigParam[int]
}

func (ip IntParam) Get(bind bool) int {
	if bind {
		ip.bind()
	}
	return viper.GetInt(ip.envar)
}

func NewIntParam(envar string, defaultValue int) IConfigParam[int] {
	return IntParam{GenericEnvConfigParam: NewGenericConfigParam(envar, defaultValue)}
}

type UIntParam struct {
	GenericEnvConfigParam[uint64]
}

func (ip UIntParam) Get(bind bool) uint64 {
	if bind {
		ip.bind()
	}
	return viper.GetUint64(ip.envar)
}

func NewUIntParam(envar string, defaultValue uint64) IConfigParam[uint64] {
	return UIntParam{GenericEnvConfigParam: NewGenericConfigParam(envar, defaultValue)}
}

type Int64Param struct {
	GenericEnvConfigParam[int64]
}

func (ip Int64Param) Get(bind bool) int64 {
	if bind {
		ip.bind()
	}
	return viper.GetInt64(ip.envar)
}

func NewInt64Param(envar string, defaultValue int64) IConfigParam[int64] {
	return Int64Param{GenericEnvConfigParam: NewGenericConfigParam(envar, defaultValue)}
}

type BoolParam struct {
	GenericEnvConfigParam[bool]
}

func (ip BoolParam) Get(bind bool) bool {
	if bind {
		ip.bind()
	}
	return viper.GetBool(ip.envar)
}

func NewBoolParam(envar string, defaultValue bool) IConfigParam[bool] {
	return BoolParam{GenericEnvConfigParam: NewGenericConfigParam(envar, defaultValue)}
}

func (cp GenericEnvConfigParam[T]) bind() {
	_ = viper.BindEnv(cp.envar)
}

func (cp GenericEnvConfigParam[T]) establishDefault() {
	if cp.hasDefault {
		viper.SetDefault(cp.envar, cp.defaultValue)
	}
}

func NewGenericConfigParamWithoutDefault(envar string) GenericEnvConfigParam[string] {
	param := GenericEnvConfigParam[string]{envar: envar, hasDefault: false}
	return param
}

func NewGenericConfigParam[T any](envar string, defaultValue T) GenericEnvConfigParam[T] {
	param := GenericEnvConfigParam[T]{envar: envar, defaultValue: defaultValue, hasDefault: true}
	param.establishDefault()
	return param
}

/* the runtime configuration parameters for our service */

type PostgresqlConfig struct {
	DbHost       string
	DbHostReader string
	DbPort       string
	DbUser       string
	DbPass       string
	DbName       string
	PgSSLMode    string
	SSLRootCert  string

	// Testing
	// Allow test-time injection of a gorm db for in memory tests
	TestDatabaseConn gorm.Dialector

	SQLMaxConnLifetimeMinutes time.Duration
	SQLMaxOpenConns           int
	SQLMaxIdleConns           int
	SQLStatTrackingDepth      int
	SQLWaitCountLimit         int64
}

func NewPostgresqlConfigFromEnv(prefix string, loadFromEnvironment bool) PostgresqlConfig {
	dbConfig := PostgresqlConfig{
		DbHost:                    NewConfigParam(prefix + "PGHOST").Get(loadFromEnvironment),
		DbHostReader:              NewConfigParam(prefix + "PGHOST_READER").Get(loadFromEnvironment),
		DbPort:                    NewConfigParam(prefix + "PGPORT").Get(loadFromEnvironment),
		DbUser:                    NewConfigParam(prefix + "PGUSER").Get(loadFromEnvironment),
		DbPass:                    NewConfigParam(prefix + "PGPASSWORD").Get(loadFromEnvironment),
		DbName:                    NewConfigParam(prefix + "PGDATABASE").Get(loadFromEnvironment),
		PgSSLMode:                 NewConfigParam(prefix + "PGSSLMODE").Get(loadFromEnvironment),
		SSLRootCert:               NewConfigParam(prefix + "PGSSLROOTCERT").Get(loadFromEnvironment),
		SQLMaxOpenConns:           NewIntParam("SQL_MAX_OPEN_CONNS", 3).Get(loadFromEnvironment),
		SQLMaxIdleConns:           NewIntParam("SQL_MAX_IDLE_CONNS", 3).Get(loadFromEnvironment),
		SQLMaxConnLifetimeMinutes: NewDurationConfigParam("SQL_MAX_CONN_LIFETIME_MINUTES", 10, time.Minute).Get(loadFromEnvironment),
		SQLStatTrackingDepth:      NewIntParam("SQL_STAT_TRACKING_DEPTH", 2).Get(loadFromEnvironment),
		SQLWaitCountLimit:         NewInt64Param("SQL_WAIT_COUNT_LIMIT", 25000).Get(loadFromEnvironment),
	}

	return dbConfig
}

type Config struct {
	ServerAddr string

	// SQL
	PostgresqlConfig PostgresqlConfig

	// REDIS
	NoSQLConnection     string
	NoSQLAuthToken      string
	NoSQLTimeout        time.Duration
	NoSQLWriteTimeout   time.Duration
	NoSQLReadTimeout    time.Duration
	NoSQLIdleTimeout    time.Duration
	NoSQLDialTimeout    time.Duration
	NoSQLPoolTimeout    time.Duration
	NoSQLPoolSize       int
	NoSQLMinIdleConns   int
	NoSQLRouteRandomly  bool
	NoSQLRouteByLatency bool
	NoSQLTLS            bool
	NoSQLTTL            time.Duration

	// METRICS
	WavefrontEndpoint            string
	WavefrontToken               string
	WavefrontPort                int
	WavefrontTracePort           int
	WavefrontHistogramPort       int
	WavefrontHistogramMinutePort int
	WavefrontHistogramHourPort   int
	WavefrontHistogramDayPort    int

	MetricPrefix      string
	MetricInterval    time.Duration
	MetricSamplerRate uint64
	BaseDimensions    map[string]string

	// MISC
	CbdEnvName        string
	Region            string
	ServicePrefix     string
	ProfilingEnabled  bool
	EnvCategory       string
	AuditingTimeDelta time.Duration // delta in seconds

	// HTTP
	HttpReadTimeout       time.Duration
	HttpWriteTimeout      time.Duration
	HttpHandleTimeout     time.Duration
	RbacHTTPHandleTimeout time.Duration
	RetryAfterSeconds     int

	// HTTP HEADERS
	RequestIDHeader string
	UserAgent       string
	AppName         string
	AppVersion      string
}

func (cfg *Config) IsLocalEnv() bool {
	return cfg.EnvCategory == "local"
}

func (cfg *Config) IsDevEnv() bool {
	return cfg.EnvCategory == "dev"
}

func newConfigFromEnv(loadFromEnvironment bool) (*Config, error) {
	var err error
	profilerEnabled := false
	if profilerFlag != nil {
		profilerEnabled = *profilerFlag
	}

	cfg := Config{
		ServerAddr: NewConfigParamWithDefault("SERVER_ADDR", "[::]:5004").Get(loadFromEnvironment),

		// SQL
		PostgresqlConfig: NewPostgresqlConfigFromEnv("", loadFromEnvironment),

		// REDIS
		NoSQLConnection:     NewConfigParam("REDIS_ADDRESS").Get(loadFromEnvironment),
		NoSQLAuthToken:      NewConfigParam("REDIS_AUTH_TOKEN").Get(loadFromEnvironment),
		NoSQLTimeout:        NewDurationConfigParam("NOSQL_TIMEOUT_MS", 500, time.Millisecond).Get(loadFromEnvironment),
		NoSQLWriteTimeout:   NewDurationConfigParam("NOSQL_WRITE_TIMEOUT_MS", 2000, time.Millisecond).Get(loadFromEnvironment),
		NoSQLReadTimeout:    NewDurationConfigParam("NOSQL_READ_TIMEOUT_MS", 2000, time.Millisecond).Get(loadFromEnvironment),
		NoSQLIdleTimeout:    NewDurationConfigParam("NOSQL_DIAL_TIMEOUT_MS", 3000, time.Millisecond).Get(loadFromEnvironment),
		NoSQLPoolTimeout:    NewDurationConfigParam("NOSQL_POOL_TIMEOUT_MS", 3000, time.Millisecond).Get(loadFromEnvironment),
		NoSQLPoolSize:       NewIntParam("NOSQL_POOL_SIZE", 5).Get(loadFromEnvironment),
		NoSQLMinIdleConns:   NewIntParam("NOSQL_MIN_IDLE_CONNS", 3).Get(loadFromEnvironment),
		NoSQLRouteRandomly:  NewBoolParam("NOSQL_ROUTE_RANDOMLY", false).Get(loadFromEnvironment),
		NoSQLRouteByLatency: NewBoolParam("NOSQL_ROUTE_BY_LATENCY", true).Get(loadFromEnvironment),
		NoSQLTLS:            NewBoolParam("NOSQL_TLS", true).Get(loadFromEnvironment),
		NoSQLTTL:            NewDurationConfigParam("NOSQL_TTL_DURATION", 45, time.Minute).Get(loadFromEnvironment),

		// METRICS
		WavefrontEndpoint:  NewConfigParam("WAVEFRONT_ENDPOINT").Get(loadFromEnvironment),
		WavefrontPort:      NewIntParam("WAVEFRONT_PORT", 2878).Get(loadFromEnvironment),
		WavefrontTracePort: NewIntParam("WAVEFRONT_TRACE_PORT", 30000).Get(loadFromEnvironment),

		MetricPrefix:      "psc.ems-device-hints",
		MetricInterval:    NewDurationConfigParam("METRIC_INTERVAL_SECONDS", 10, time.Second).Get(loadFromEnvironment),
		MetricSamplerRate: NewUIntParam("METRIC_SAMPLER_RATE", 5).Get(loadFromEnvironment),

		// MISC
		CbdEnvName: NewConfigParam("CBDENVNAME").Get(loadFromEnvironment),

		ServicePrefix:     NewConfigParamWithDefault("SERVICE_PREFIX", "hinting").Get(loadFromEnvironment),
		ProfilingEnabled:  profilerEnabled,
		EnvCategory:       NewConfigParamWithDefault("ENV_CATEGORY", "local").Get(loadFromEnvironment),
		AuditingTimeDelta: NewDurationConfigParam("AUDITING_TIME_DELTA_SECONDS", 5, time.Second).Get(loadFromEnvironment),

		// HTTP
		HttpReadTimeout:       NewDurationConfigParam("HTTP_READ_TIMEOUT_MS", 5000, time.Millisecond).Get(loadFromEnvironment),
		HttpWriteTimeout:      NewDurationConfigParam("HTTP_WRITE_TIMEOUT_MS", 5000, time.Millisecond).Get(loadFromEnvironment),
		HttpHandleTimeout:     NewDurationConfigParam("HTTP_HANDLE_TIMEOUT_MS", 5000, time.Millisecond).Get(loadFromEnvironment),
		RbacHTTPHandleTimeout: NewDurationConfigParam("RBAC_HTTP_TIMEOUT_S", 4*time.Second, time.Second).Get(loadFromEnvironment),
		// HTTP HEADERS
		RequestIDHeader:   "Wf-Ot-Traceid",
		RetryAfterSeconds: NewIntParam("RETRY_AFTER_SECONDS", 5).Get(loadFromEnvironment),
		AppName:           NewConfigParamWithDefault("APP_NAME", "ems-device-hints").Get(loadFromEnvironment),
		AppVersion:        NewConfigParamWithDefault("COMMIT_SHA", "1.0.0").Get(loadFromEnvironment),
	}

	cfg.UserAgent = cfg.AppName + "/" + cfg.AppVersion
	hostName, err := os.Hostname()
	if err != nil {
		return nil, err
	}
	cfg.BaseDimensions = map[string]string{
		"product":     "psc",
		"application": "ems-device-hints",
		"service":     "ems-device-hints",
		"env":         cfg.CbdEnvName,
		"instance":    hostName,
	}
	return &cfg, err
}
