// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package logger

import (
	"context"
	"fmt"
	"net/http"
)

// Level represents the level of a log message.
type Level int

//goland:noinspection GoUnusedConst
const (
	// FatalLevel is a log level for severe error events that will presumably lead the application to abort.
	FatalLevel Level = iota // 0
	// ErrorLevel is a log level for events that might still allow the application to continue running.
	ErrorLevel
	// WarnLevel is a log level for potentially harmful situations.
	WarnLevel
	// InfoLevel is a log level for informational messages that highlight the progress of the application at coarse-grained level.
	InfoLevel
	// DebugLevel is a log level for fine-grained informational events that are most useful to debug an application.
	DebugLevel
)

type requestLoggerKey int

// ContextWithRequestLogger returns a context with the given ids stored
func ContextWithRequestLogger(ctx context.Context, l Logger) context.Context {
	var key requestLoggerKey
	return context.WithValue(ctx, key, l)
}

// RequestLoggerFromContext gets ids stored in a context
func RequestLoggerFromContext(ctx context.Context) Logger {
	var key requestLoggerKey
	l, ok := ctx.Value(key).(Logger)
	if !ok || l == nil {
		l = BaseLogger()
		l.Error("failed to fetch logger from context")
	}
	return l
}

// RequestMessage logs a message with the request logger at the given log level
func RequestMessage(r *http.Request, lvl Level, m interface{}) {
	var msg string
	switch v := m.(type) {
	case string:
		msg = v
	case error:
		msg = v.Error()
	default:
		msg = fmt.Sprintf("unable to log message type: %T", v)
	}

	l := RequestLoggerFromContext(r.Context())
	switch lvl {
	case ErrorLevel:
		l.Error(msg)
	case WarnLevel:
		l.Warn(msg)
	case InfoLevel:
		l.Info(msg)
	case DebugLevel:
		l.Debug(msg)
	default:
		l.Error(fmt.Sprintf("unable to log message at level %v", lvl))
	}
}

// RequestError logs an error message on the request logger
//
//goland:noinspection GoUnusedExportedFunction
func RequestError(r *http.Request, msg interface{}) {
	RequestMessage(r, ErrorLevel, msg)
}

// RequestWarn logs an warn message on the request logger
//
//goland:noinspection GoUnusedExportedFunction
func RequestWarn(r *http.Request, msg interface{}) {
	RequestMessage(r, WarnLevel, msg)
}

// RequestInfo logs an info message on the request logger
//
//goland:noinspection GoUnusedExportedFunction
func RequestInfo(r *http.Request, msg interface{}) {
	RequestMessage(r, InfoLevel, msg)
}

// RequestDebug logs an debug message on the request logger
//
//goland:noinspection GoUnusedExportedFunction
func RequestDebug(r *http.Request, msg interface{}) {
	RequestMessage(r, DebugLevel, msg)
}
